@@ -20,6 +20,7 @@
 #'                   in order to reduce the number of optimization parameters, and potentially
 #'                   improving computational performance.
 #' @param spline.df Number of degrees of freedom (per running variable) used for spline computation.
+#' @param try.elnet.for.sigma.sq Whether an elastic net on a spline basis should be used for estimating sigma^2.
 #' @param optimizer Which optimizer to use? Mosek is a commercial solver, but free
 #'                  academic licenses are available. Needs to be installed separately.
 #'                  ECOS is an open-source interior-point solver for conic problems,
@@ -94,11 +95,13 @@ optrdd = function(X,
                   use.homoskedatic.variance = FALSE,
                   use.spline = TRUE,
                   spline.df = NULL,
+                  try.elnet.for.sigma.sq = FALSE,
                   optimizer = c("auto", "mosek", "ECOS", "quadprog", "SCS"),
                   verbose = TRUE) {
 
     n = length(W)
     if (class(W) == "logical") W = as.numeric(W)
+    if (!all(W %in% c(0, 1))) stop("The treatment assignment W must be binary.")
     if (!is.null(Y) & (length(Y) != n)) { stop("Y and W must have same length.") }
     if (is.null(dim(X))) { X = matrix(X, ncol = 1) }
     if (nrow(X) != n) { stop("The number of rows of X and the length of W must match") }
@@ -114,13 +117,42 @@ optrdd = function(X,
         ((max(X[W==0, 1]) <= min(X[W==1, 1])) || (max(X[W==1, 1]) <= min(X[W==0, 1])))
     
     mosek_available = requireNamespace("Rmosek", quietly = TRUE)
-    
+
+    # Naive initialization for sigma.sq if needed
+    if (is.null(sigma.sq)) {
+        if (is.null(Y)) {
+            warning("Setting noise level to 1 as default...")
+            sigma.sq = 1
+        } else {
+            Y.hat = stats::predict(stats::lm(Y ~ X * W))
+            sigma.sq = mean((Y - Y.hat)^2) * length(W) / (length(W) - 2 - 2 * nvar)
+            if (try.elnet.for.sigma.sq){
+                if(ncol(X) > 1) {
+                    stop("Elastic net for sigma squared not implemented with more than 1 running variable.")
+                }
+                linear.params = 1 + 2 * ncol(X)
+                elnet.df = 7
+                ridge.mat = cbind(W, X, W * X, matrix(0, length(W), 2 * elnet.df))
+                ridge.mat[W==0, linear.params + 1:elnet.df] = splines::ns(X[W==0,], df = elnet.df)
+                ridge.mat[W==1, linear.params + elnet.df + 1:elnet.df] = splines::ns(X[W==1,], df = elnet.df)
+                elnet = glmnet::cv.glmnet(ridge.mat, Y,
+                                          penalty.factor = c(rep(0, linear.params),
+                                                             rep(1, 2 * elnet.df)),
+                                          keep = TRUE, alpha = 0.5)
+                elnet.hat = elnet$fit.preval[,!is.na(colSums(elnet$fit.preval)),drop=FALSE][, elnet$lambda == elnet$lambda.1se]
+                sigma.sq.elnet = mean((elnet.hat - Y)^2)
+                sigma.sq = min(sigma.sq, sigma.sq.elnet)
+            }
+        }
+    }
+
     optimizer = match.arg(optimizer)
     if (optimizer == "auto") {
         if (nvar == 1 &&
             use.spline &&
             (univariate.monotone || !cate.at.pt) &&
-            length(unique(c(X))) <= 100) {
+            length(unique(c(X))) <= 100 &&
+            max.second.derivative / sigma.sq <= 4) {
             optimizer = "quadprog"
         } else {
             if (mosek_available) {
@@ -130,8 +162,7 @@ optrdd = function(X,
             }
         }
     }
-        
-    
+
     if (optimizer == "mosek") {
         if (!mosek_available) {
             optimizer = "ECOS"
@@ -149,19 +180,14 @@ optrdd = function(X,
         #         warning(paste("The mosek optimizer is not installed; using quadprog instead."))
         #     }
         # }
-    }  
-    
-    # Naive initialization for sigma.sq if needed
-    if (is.null(sigma.sq)) {
-        if (is.null(Y)) {
-            warning("Setting noise level to 1 as default...")
-            sigma.sq = 1
-        } else {
-            Y.hat = stats::predict(stats::lm(Y ~ X * W))
-            sigma.sq = mean((Y - Y.hat)^2) * length(W) / (length(W) - 2 - 2 * nvar)
-        }
     }
-    
+
+    if (optimizer == "SCS") {
+        warning(paste("SCS is a fast an free optimizer, but doesn't solve the problem exactly.",
+                      "Resulting confidence intervals may be needlessly long.",
+                      "It is recommended to also try MOSEK."))
+    }
+
     # Create discrete grid on which to optimize, and assign each training example
     # to a cell.
     if (nvar == 1) {
@@ -415,6 +441,13 @@ optrdd = function(X,
         )
         cvx.problem = CVXR::Problem(CVXR::Minimize(objective), contraints)
         cvx.output = solve(cvx.problem, solver = optimizer, verbose = verbose)
+        
+        if (cvx.output$status != "optimal") {
+            warning(paste0("CVXR returned with status: ",
+                           cvx.output$status,
+                           ". For better results, try another optimizer (MOSEK is recommended)."))
+        }
+        
         result = cvx.output$getValue(xx)
         gamma.0[realized.idx.0] = - result[1:num.realized.0] / sigma.sq / 2
         gamma.1[realized.idx.1] = - result[num.realized.0 + 1:num.realized.1] / sigma.sq / 2
@@ -453,6 +486,12 @@ optrdd = function(X,
             mosek.out = Rmosek::mosek(mosek.problem, opts=list(verbose=0))
         }
         
+        if (mosek.out$response$code != 0) {
+            warning(paste("MOSEK returned with status",
+                          mosek.out$response$msg,
+                          "For better results, try another optimizer."))
+        }
+        
         # We now also need to re-adjust for "natural" scaling
         gamma.0[realized.idx.0] = - mosek.out$sol$itr$xx[1:num.realized.0] / sigma.sq / 2 /
             sqrt(Dmat.diagonal[1:num.realized.0])
@@ -486,12 +525,21 @@ optrdd = function(X,
         if (use.homoskedatic.variance) {
             se.hat.tau = sqrt(sum(gamma^2 * sigma.sq))
         } else {
-            # A heteroskedaticity-robust variance estimate
-            regr.df = data.frame(X=X, W=W, Y=Y)
-            Y.fit = stats::lm(Y ~ X * W, data = regr.df)
-            Y.resid.sq = (Y - stats::predict(Y.fit))^2 * length(W) /
-                (length(W) - 2 * (1 + nvar))
-            se.hat.tau = sqrt(sum(Y.resid.sq * gamma^2))
+            # A heteroskedaticity-robust variance estimate.
+            # Weight the regression towards areas used in estimation, so we are not
+            # too vulnerable to curvature effects far from the boundary.
+            regr.df = data.frame(X=X, W=W, Y=Y, gamma.sq = gamma^2)
+            regr.df = regr.df[regr.df$gamma.sq != 0,]
+            Y.fit = stats::lm(Y ~ X * W, data = regr.df, weights = regr.df$gamma.sq)
+            self.influence = stats::lm.influence(Y.fit)$hat
+            Y.resid.sq = (regr.df$Y - stats::predict(Y.fit))^2
+            se.hat.tau = sqrt(sum(Y.resid.sq * regr.df$gamma.sq / (1 - self.influence)))
+
+            if (!try.elnet.for.sigma.sq & se.hat.tau^2 < 0.8 * sum(regr.df$gamma.sq) * sigma.sq) {
+                warning(paste("Implicit estimate of sigma^2 may be too pessimistic,",
+                              "resulting in valid but needlessly long confidence intervals.",
+                              "Try the option `try.elnet.for.sigma.sq = TRUE` for potentially improved performance."))
+            }
         }
         
         # Confidence intervals that account for both bias and variance
